//---------------------------------------------------------------------------

#ifdef __BORLANDC__
#	pragma hdrstop
#endif

#include "ComPort.h"
#include <stdlib.h>
#include "msports.h"

//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#	pragma package(smart_init)
#endif

std::vector<struct ComPort::S_COM_PORT> ComPort::ports;

ComPort comPort;

int ComPort::EnumeratePorts(void)
{
    OSVERSIONINFO osvinfo;
    memset(&osvinfo, 0, sizeof(osvinfo));
    osvinfo.dwOSVersionInfoSize=sizeof(osvinfo);

    GetVersionEx(&osvinfo);

    switch (osvinfo.dwPlatformId)
    {
    case VER_PLATFORM_WIN32_WINDOWS:
		return W9xEnumeratePorts();
	case VER_PLATFORM_WIN32_NT:
		if (osvinfo.dwMajorVersion > 4)
            return W2kEnumeratePorts();
        else
			return ERR_UNSUPPORTED_OS;
    default:
        return ERR_UNSUPPORTED_OS;
    }
}

enum ComPort::E_ERR ComPort::W9xEnumeratePorts(void)
{
	return ScanEnumBranch("ENUM");
}

enum ComPort::E_ERR ComPort::W2kEnumeratePorts(void)
{
    return ScanEnumBranch("SYSTEM\\CURRENTCONTROLSET\\ENUM");
}

std::string ComPort::GetRegistryString(const HKEY hKey, const std::string& name)
{
    DWORD type;
    DWORD size;
	if (RegQueryValueEx(hKey, name.c_str(), NULL, &type, NULL, &size) != ERROR_SUCCESS || type != REG_SZ && type != REG_EXPAND_SZ)
		return "";	// not found
	if (size > 0)
	{
		char* buffer = new char[size + 1];
		RegQueryValueEx(hKey, name.c_str(), NULL, NULL, (BYTE*) buffer, &size);
		buffer[size] = 0;
		std::string result(buffer);
		delete [] buffer;
		return result;
	}
	return "";		// empty
}

std::string ComPort::OpenRegistrySubkey(const HKEY hKey,
	DWORD index, REGSAM regsam,PHKEY phkResult)
{
	FILETIME filetime;
	DWORD size;
	DWORD MaxSubKeyLen = 0;
	DWORD SubKeys = 0;
	DWORD status = RegQueryInfoKey(hKey, NULL, NULL, NULL,
		&SubKeys,&MaxSubKeyLen,NULL,NULL,
		NULL,NULL,NULL,NULL);
	//if((status == ERROR_SUCCESS) && SubKeys)
	if (status != ERROR_SUCCESS)
	{
		/** \todo RegQueryInfoKey returns 5 not allowing to get maximum subkey size in ENUM branch */
		MaxSubKeyLen = 512;
	}

	size = MaxSubKeyLen;
	char* buffer = new char[size + 1];
	RegEnumKeyEx(hKey, index, buffer, &size, 0, NULL, NULL, &filetime);
	buffer[size] = 0;
	std::string result(buffer);
	delete [] buffer;
	if (RegOpenKeyEx(hKey, result.c_str(), 0, regsam, phkResult) == ERROR_SUCCESS)
		return result;
	return "";		// empty
}

enum ComPort::E_ERR ComPort::ScanEnumBranch(std::string path)
{
	static const char lpstrPortsClass[] = "PORTS";
	static const char lpstrPortsClassGUID[] = "{4D36E978-E325-11CE-BFC1-08002BE10318}";
	char lpClass[sizeof(lpstrPortsClass)];
	DWORD  cbClass;
	char lpClassGUID[sizeof(lpstrPortsClassGUID)];
	DWORD  cbClassGUID;

	ports.clear();
	struct S_COM_PORT port;

	HKEY hkEnum;
	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, path.c_str(), 0, KEY_ENUMERATE_SUB_KEYS, &hkEnum))
		return ERR_REGISTRY_READ;

	for (DWORD dwIndex1=0; ; ++dwIndex1)
	{
		HKEY hkLevel1;
		port.technology = OpenRegistrySubkey(hkEnum, dwIndex1, KEY_ENUMERATE_SUB_KEYS, &hkLevel1);
		if (port.technology == "")
		{
			RegCloseKey(hkEnum);
			return ERR_NONE;
		}

		for (DWORD dwIndex2=0; ; ++dwIndex2)
		{
			HKEY hkLevel2;
			std::string tmp = OpenRegistrySubkey(hkLevel1,
				dwIndex2, KEY_ENUMERATE_SUB_KEYS, &hkLevel2);
			if (tmp == "")
			{
				RegCloseKey(hkLevel2);
				break;
			}

			for (DWORD dwIndex3=0; ; ++dwIndex3)
            {
				HKEY hkLevel3;
				tmp = OpenRegistrySubkey(hkLevel2, dwIndex3, KEY_READ, &hkLevel3);
				if (tmp == "")
				{
					RegCloseKey(hkLevel3);
					break;
				}
				// search for "CLASS" or "CLASSGUID"
				cbClass=sizeof(lpClass);
				if (RegQueryValueEx(hkLevel3,TEXT("CLASS"),NULL,NULL,
									(LPBYTE)lpClass,&cbClass) != ERROR_SUCCESS ||
					strcmp(lpClass,lpstrPortsClass))
				{
					cbClassGUID=sizeof(lpClassGUID);
					if (RegQueryValueEx(hkLevel3,TEXT("CLASSGUID"),NULL,NULL,
										(LPBYTE)lpClassGUID,&cbClassGUID) != ERROR_SUCCESS ||
						strcmp(lpClassGUID,lpstrPortsClassGUID))
					{
						RegCloseKey(hkLevel3);
						continue;
					}
				}

				port.name = GetRegistryString(hkLevel3, "PORTNAME");
				if (port.name == "")
				{
					HKEY hkDevParameters;
					if (RegOpenKeyEx(hkLevel3, "DEVICE PARAMETERS", 0, KEY_READ,
						&hkDevParameters) == ERROR_SUCCESS)
					{
						port.name = GetRegistryString(hkDevParameters, "PORTNAME");
						RegCloseKey(hkDevParameters);
					}
				}
                if (port.name.find("COM", 0) != std::string::npos)
                {
					port.description = GetRegistryString(hkLevel3, "FRIENDLYNAME");
					ports.push_back(port);
                }
				RegCloseKey(hkLevel3);
			}
			RegCloseKey(hkLevel2);
		}
        RegCloseKey(hkLevel1);
    }
}

#if 0
int ComPort::EnumeratePortsAlt(void)
{
	HKEY hKey;
	std::string path = "Hardware\\DeviceMap\\SerialComm";
	ports.clear();
	HKEY hkSerialComm;
	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, path.c_str(), 0, KEY_QUERY_VALUE, &hkSerialComm) != ERROR_SUCCESS)
		return ERR_REGISTRY_READ;

	char* pValNameBuff;
	char* pValueBuff;

	DWORD dwValCount, dwMaxCharValNameLen, dwMaxWideValueSize;
	if (RegQueryInfoKey(hkSerialComm, NULL, NULL, NULL, NULL, NULL, NULL,
		&dwValCount, &dwMaxCharValNameLen, &dwMaxWideValueSize,	NULL, NULL) != ERROR_SUCCESS)
	{
		RegCloseKey(hkSerialComm);
		return ERR_REGISTRY_READ;
	}

	const dwMaxCharValNameSize = dwMaxCharValNameLen + 1;
	pValNameBuff = new char[dwMaxCharValNameSize];
	if (!pValNameBuff)
	{
		RegCloseKey(hkSerialComm);
		return ERR_NO_MEMORY;
	}
	// size is always returned in bytes needed to hold the UNICODE
	const dwMaxByteValueSize = (dwMaxWideValueSize/2)*(sizeof (TCHAR));
	pValueBuff = new char[dwMaxByteValueSize];
	if (!pValueBuff)
	{
		delete [] pValNameBuff;
		RegCloseKey(hkSerialComm);
		return ERR_NO_MEMORY;
	}
	for (DWORD dwIndex = 0; dwIndex < dwValCount; ++dwIndex)
	{
		DWORD dwCharValNameSize = dwMaxCharValNameSize;
		DWORD dwByteValueSize = dwMaxByteValueSize;
		DWORD dwType;
		LONG nRes = RegEnumValue(hkSerialComm, dwIndex, pValNameBuff,
			&dwCharValNameSize, NULL, &dwType, (LPBYTE)pValueBuff, &dwByteValueSize);
		if (nRes != ERROR_SUCCESS)
		{
			delete [] pValNameBuff;
			delete [] pValueBuff;
			RegCloseKey(hkSerialComm);
			return ERR_NO_MEMORY;
		}
		if (dwType != REG_SZ)
		{
			continue;
		}
		struct S_COM_PORT port;
		port.name = pValueBuff;
		ports.push_back(port);
	}
	delete [] pValNameBuff;
	delete [] pValueBuff;
	return ERR_NONE;
}
#endif

int ComPort::EnumeratePortsAlt(void)
{
	ports.clear();

	HKEY hkCommMap;
	if (ERROR_SUCCESS != RegOpenKeyEx(
		HKEY_LOCAL_MACHINE,
		TEXT("HARDWARE\\DEVICEMAP\\SERIALCOMM"),
		0,
		KEY_QUERY_VALUE,
		&hkCommMap))
	{
        return ERR_REGISTRY_READ;
	}

	void* pValNameBuff = 0;
	void* pValueBuff = 0;
	__try
	{
		DWORD dwValCount, dwMaxCharValNameLen, dwMaxWideValueSize;
		if (ERROR_SUCCESS != RegQueryInfoKey(
			hkCommMap,
			NULL, NULL, NULL, NULL, NULL, NULL,
			&dwValCount,
			&dwMaxCharValNameLen,
			&dwMaxWideValueSize,
			NULL, NULL))
		{
	    	RegCloseKey(hkCommMap);
    		return ERR_REGISTRY_READ;
		}

		// The max value name size is returned in TCHARs not including
		// the terminating null character.
		const dwMaxCharValNameSize = dwMaxCharValNameLen + 1;
		pValNameBuff = malloc(dwMaxCharValNameSize*(sizeof (TCHAR)));
		if (!pValNameBuff)
		{
            return -1;
		}

		// The max value size is returned in bytes needed to hold the UNICODE
		// strings including terminating null character regardless of the
		// function type (ANSI or UNICODE).
		const dwMaxByteValueSize = (dwMaxWideValueSize/2)*(sizeof (TCHAR));
		pValueBuff = malloc(dwMaxByteValueSize);
		if (!pValueBuff)
		{
            return -1;
		}

		for (DWORD dwIndex = 0; dwIndex < dwValCount; ++dwIndex)
		{
			DWORD dwCharValNameSize = dwMaxCharValNameSize;
			DWORD dwByteValueSize = dwMaxByteValueSize;
			DWORD dwType;
			LONG nRes = RegEnumValue(
				hkCommMap,
				dwIndex,
				(LPTSTR)pValNameBuff,
				&dwCharValNameSize,
				NULL,
				&dwType,
				(LPBYTE)pValueBuff,
				&dwByteValueSize);
			if (nRes != ERROR_SUCCESS)
			{
				break;
			}
			if (dwType != REG_SZ)
			{
				continue;
			}

    		struct S_COM_PORT port;
	    	port.name = (char*)pValueBuff;
    		ports.push_back(port);
		}
	}
	__finally
	{
		free(pValueBuff);
		free(pValNameBuff);
		RegCloseKey(hkCommMap);
	}
    return 0;
}

int ComPort::ReleasePortNumbers(std::vector<int> &ports)
{
	HANDLE hInstance = LoadLibrary("msports.dll");
	if (!hInstance)
		return -1;
	typedef LONG WINAPI (*pComDBOpen) (PHCOMDB PHComDB);
	typedef LONG WINAPI (*pComDBClose) (HCOMDB PHComDB);
	typedef LONG WINAPI (*pComDBReleasePort) (HCOMDB   HComDB, DWORD ComNumber);

	pComDBOpen pComDBOpenFn = (pComDBOpen)		GetProcAddress(hInstance, "ComDBOpen");
	pComDBClose pComDBCloseFn = (pComDBClose)		GetProcAddress(hInstance, "ComDBClose");
	pComDBReleasePort pComDBReleasePortFn = (pComDBReleasePort)   GetProcAddress(hInstance, "ComDBReleasePort");
	if (pComDBOpenFn == NULL ||
		pComDBCloseFn == NULL ||
		pComDBReleasePortFn == NULL)
	{
		FreeLibrary(hInstance);
		return -2;
	}

	//update COM port database
	HCOMDB comDB;
	LONG status;
	if ((status = pComDBOpenFn(&comDB)) != ERROR_SUCCESS)
	{
		FreeLibrary(hInstance);
		return ERR_COMDB;
	}
	for (unsigned int i=0; i<ports.size(); i++)
	{
		if ((status = pComDBReleasePortFn(comDB, ports[i])) != ERROR_SUCCESS)
		{
			pComDBCloseFn(comDB);
			FreeLibrary(hInstance);
			return ERR_COMDBRELEASE;
		}
	}
	pComDBCloseFn(comDB);
	(void)status;
	FreeLibrary(hInstance);
	return ERR_NONE;
}

ComPort::ComPort():
	handle(INVALID_HANDLE_VALUE)
{
}

int ComPort::Open(std::string name)
{
	Close();
	return ERR_NONE;
}

int ComPort::Close(void)
{
	if (handle != INVALID_HANDLE_VALUE)
	{
		CloseHandle(handle);
		handle = INVALID_HANDLE_VALUE;
	}
	return ERR_NONE;
}

BOOL ReadComm(HANDLE hCommDev, LPVOID lpBuffer, LPDWORD lpNumberOfBytesRead, DWORD BufSize)
{
	*lpNumberOfBytesRead = 0;
	return ReadFile(hCommDev, lpBuffer, BufSize, lpNumberOfBytesRead, NULL);
}

BOOL WriteComm(HANDLE hCommDev, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite)
{
	DWORD nNumberOfBytesWritten;
	BOOL writeResult = WriteFile(hCommDev, lpBuffer, nNumberOfBytesToWrite, &nNumberOfBytesWritten, NULL);
	if (nNumberOfBytesWritten != nNumberOfBytesToWrite)
	{
		return FALSE;
	}
	return writeResult;
}

