//---------------------------------------------------------------------------

#ifndef DSO183H
#define DSO183H
//---------------------------------------------------------------------------

#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include "WorkerThread.h"

enum { CHANNELS = 3 };

class TestScope
{
private:
	bool bManualTrigger;
	bool bDeviceOpened;
	bool bRunning;
	bool bStopRequest;
	enum E_TRIGGER_MODE mode;
	struct {
		float fDivX;
		float fDivY[CHANNELS];
		bool channelEnabled[CHANNELS];
		int samples;
		int triggerval;
		int triggerpos;
		int triggerChannel;
		enum E_TRIGGER_TYPE triggerType;
	} settings;

	WorkerThread *workerThread;	

	typedef void (*CallbackLog)(char*);
	typedef void (*CallbackData)(short*, int);
	typedef void (*CallbackConnect)(int, char*);
	typedef void (*CallbackTrigger)(void);
	void log(char *lpData, ...);

	short buffer[CHANNELS * 1024 * 1024];

	int lastTriggerDistance;

	int TestTrigger(short* output_buffer, int len);

public:
	TestScope();
	~TestScope();
	int Connect(void);
	int Disconnect(void);
	int RecStart(void);				///< start recording
	int RecStop(void);              ///< stop recording
	CallbackLog callbackLog;
	CallbackData callbackData;
	CallbackConnect callbackConnect;
	CallbackTrigger callbackTrigger;
	void ConnectCb(int state, char* msg) {
		if (callbackConnect)
			callbackConnect(state, msg);
	}
	/** \brief Request manual triggering
		\return 0 on success
	*/
	int ManualTrigger(void);
	int Run(bool state);
	int SetTriggerMode(enum E_TRIGGER_MODE);	
	int SetTriggerType(int channel, enum E_TRIGGER_TYPE);
	int EnableChannel(int channel, bool state);
	int SetCouplingType(int channel, enum E_COUPLING_TYPE);
	/** \brief Set capture buffer size
		\param size number of samples
	*/
	int SetBufferSize(int size);
	int SetSensitivity(int channel, int id);
	int SetSampling(int id);
	int SetTriggerLevel(int level);
	int SetTriggerPosition(int pos);

	void handle(void);
};
extern TestScope testScope;


#endif
